// THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
// ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
// THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
// PARTICULAR PURPOSE.
//
// Copyright (C) 1998  Microsoft Corporation.  All Rights Reserved.
//
// Module:
//      receiver.cpp
//
// Abstract: 
//      Implement receive-side functionality.  The reciever can accept up to 
//      MAXCLIENT connections from senders (no limit for UDP).  The receiver 
//      uses WSAEventSelect to handle the potential multiple connections as well 
//      as to allow the receiver to receive notification of QOS activity via the 
//      FD_QOS notification.
//
//      Additionally, the receiver can be setup to receive multicast data for a given
//      group.  This is accomplished by using the -m:Group-IP option which joins the
//      receiver to that Group-IP multicast group.
//
//      The template used when enabling QOS is specified with the commandline option
//      -q:template.  A value of default for template will result in default values
//      being entered for sending and receiving flowspec (see FillQos).  Otherwise
//      the specified template is used (again see FillQos).
//
//      QOS can be enabled in the receiver in the following manner
//      -i:b    before WSAAccept on the listening socket using WSAIoctl
//      -i:d    during WSAAccept in the condition function
//      -i:a    after WSAAccept on the accepted socket using WSAIoctl
//      -i:q    upon receipt of FD_QOS.  In order to receive FD_QOS (i.e. enable
//              the gqossp) qos is set using WSAIoctl prior to WSAEventSelect
//  
//      Receipt of RESV confirmation can be requested using the -c commandline option
//      (not yet implemented).  
//
// Author:
//      Barry R. Butterklee
//

#include "qosevent.h"


// encapsulate socket, events, buffers, and various flags and counters that
// control how/when data is sent into a structure
typedef struct _RECV_INFO
    {
    SOCKET          sd[MAXCLIENTS+1];
    WSAEVENT        hEvents[MAXCLIENTS+1];
    CHAR *          recvbuf;
    int             nTotalClients;
    } RECV_INFO;


static SOCKET g_sd;


static BOOL CreateReceivingSocket(RECV_INFO *, OPTIONS *, WSAPROTOCOL_INFO *);
static BOOL fd_accept_func(DWORD, WSANETWORKEVENTS *, RECV_INFO *, OPTIONS *);
static BOOL fd_read_func(DWORD, WSANETWORKEVENTS *, RECV_INFO *, OPTIONS *);
static BOOL fd_qos_func(DWORD, WSANETWORKEVENTS *, RECV_INFO *, OPTIONS *);
static BOOL fd_close_func(DWORD, WSANETWORKEVENTS *, RECV_INFO *, OPTIONS *);
int CALLBACK AcceptCondFunc(
    LPWSABUF, LPWSABUF, LPQOS, LPQOS, LPWSABUF, LPWSABUF, GROUP FAR *, DWORD);




// Abstract:
//      Main function to create a socket that allows clients to connect and
//      send it data.  Upto MAXCLIENTS can connect and send data.  The data
//      is read in and then tossed (the data is not used).  
//
//      When using TCP, the first socket/event in the array of sockets/events 
//      (RECV_INFO) is the listening socket and the remaining are to handle the 
//      accepted connections.  For UDP, only the first socket/event is used.
//
VOID Receiver(
    OPTIONS             *pOptions, 
    WSAPROTOCOL_INFO    *pProtocolInfo
    )
    {
    WSANETWORKEVENTS NetworkEvents;
    long lNetworkEvents;
    BOOL bProcessEventsDone;
    RECV_INFO RecvInfo;
    int i;

    printf("Receiver\n");

    // initialize data structures
    for (i=0; i<MAXCLIENTS+1; i++)
        {
        RecvInfo.sd[i] = INVALID_SOCKET;
        if (WSA_INVALID_EVENT == (RecvInfo.hEvents[i] = WSACreateEvent()))
            {
            printf("WSACreateEvent(%d): %d\n", i, WSAGetLastError());
            goto CLEANUP;
            }
        }
        
    // create the receiving/listening socket
    if (!CreateReceivingSocket(&RecvInfo, pOptions, pProtocolInfo))
        goto CLEANUP;


    // TCP sockets need to accept a connection prior to receiving data
    if (IPPROTO_TCP == pOptions->spOptions.iProtocol)
        lNetworkEvents = FD_ACCEPT;
    else
        lNetworkEvents = (FD_READ|FD_QOS|FD_CLOSE);
    if (SOCKET_ERROR == WSAEventSelect(RecvInfo.sd[0], RecvInfo.hEvents[0], lNetworkEvents))
        {
        printf("WSAEventSelect: %d\n", WSAGetLastError());
        goto CLEANUP;
        }

    bProcessEventsDone = FALSE;
    while (!bProcessEventsDone)
        {
        DWORD dwEvent;
    
        dwEvent = WSAWaitForMultipleEvents((MAXCLIENTS+1), RecvInfo.hEvents, FALSE, WSA_INFINITE, FALSE);
        switch (dwEvent)
            {
            case WSA_WAIT_FAILED:
                printf("WSAEventSelect: %d\n", WSAGetLastError());
                bProcessEventsDone = TRUE;
                break;
            case WAIT_IO_COMPLETION:
            case WSA_WAIT_TIMEOUT:
                break;

            default:
                // lets see what network activity trigged this event
                NetworkEvents.lNetworkEvents = 0;
                if (SOCKET_ERROR == WSAEnumNetworkEvents(RecvInfo.sd[dwEvent], RecvInfo.hEvents[dwEvent], &NetworkEvents))
                    {
                    printf("WSAEnumNetworkEvent: %d dwEvent %d lNetworkEvent %X\n", 
                        WSAGetLastError(), dwEvent, NetworkEvents.lNetworkEvents);
                    NetworkEvents.lNetworkEvents = 0;
                    bProcessEventsDone = TRUE;
                    }
                else
                    {
                    // it is possible that more than one notification triggered the event
                    //printf("WSAEnumNetworkEvents %X\n", NetworkEvents.lNetworkEvents);
                    if (FD_ACCEPT & NetworkEvents.lNetworkEvents)
                        bProcessEventsDone |= 
                            fd_accept_func(dwEvent, &NetworkEvents, &RecvInfo, pOptions);
                    if (FD_READ   & NetworkEvents.lNetworkEvents)
                        bProcessEventsDone |= 
                            fd_read_func(dwEvent, &NetworkEvents, &RecvInfo, pOptions);
                    if (FD_QOS    & NetworkEvents.lNetworkEvents)
                        bProcessEventsDone |= 
                            fd_qos_func(dwEvent, &NetworkEvents, &RecvInfo, pOptions);
                    if (FD_CLOSE  & NetworkEvents.lNetworkEvents)
                        bProcessEventsDone |= 
                            fd_close_func(dwEvent, &NetworkEvents, &RecvInfo, pOptions);
                    }
                break;
            }
        } // while (!bProcessEventsDone)


CLEANUP:
    for (i=0; i<MAXCLIENTS+1; i++)
        {
        WSACloseEvent(RecvInfo.hEvents[i]);
        closesocket(RecvInfo.sd[i]);
        }

    return;
    }




// Abstract:
//      Create a socket on which to receive data.  Note that this differs depending
//      upon whether TCP or UDP is used.  Also, please note that WSA_FLAG_OVERLAPPED 
//      *MUST* be specified.  Also, note that the parameters for WSASocket and the 
//      data to fill in the SOCKADDR_IN structure come from the supplied protocol 
//      info structure instead of being hardcoded.
//
//      The listening socket (and thus all accepted sockets) can be made into a QOS
//      socket if the -i:b option is set.  If -i:q is set then we actually need to
//      inform the GQOSSP that we wish to receive FD_QOS if we ask for them.  As a
//      consequence we also need to make the socket into a QOS socket before 
//      WSAEventSelect is called.  I choose to make the socket a QOS socket (if -i:q)
//      here to accomadate both TCP and UDP sockets instead of in the FD_ACCEPT 
//      handler because the FD_ACCEPT handler will not be invoked for UDP.
//
static BOOL CreateReceivingSocket(
    RECV_INFO           *pRecvInfo, 
    OPTIONS             *pOptions, 
    WSAPROTOCOL_INFO    *pProtocolInfo
    )
    {
    SOCKADDR_IN addr_srv;
    SOCKADDR_IN mult_addr;
    int         nRet;
    QOS         qos;
    QOS         *pQos;
    DWORD       cbQosLen;
    DWORD       dwSocketFlags;


    dwSocketFlags = WSA_FLAG_OVERLAPPED;

    // If we are multicast enabled set the control and data planes to leafs.
    // These flags must be set for IP multicasting.
    if (pOptions->spOptions.bMulticast)
        dwSocketFlags |= (WSA_FLAG_MULTIPOINT_C_LEAF | WSA_FLAG_MULTIPOINT_D_LEAF);

    pRecvInfo->sd[0] = WSASocket(FROM_PROTOCOL_INFO, FROM_PROTOCOL_INFO, FROM_PROTOCOL_INFO, 
            pProtocolInfo, 0, dwSocketFlags);             
    if (INVALID_SOCKET == pRecvInfo->sd[0])
        {
        printf("WSASocket %d", WSAGetLastError());
        return(FALSE);
        }

    ZeroMemory((PVOID)&addr_srv, sizeof(addr_srv));
    addr_srv.sin_family      = pProtocolInfo->iAddressFamily;
    addr_srv.sin_addr.s_addr = htonl(INADDR_ANY);
    addr_srv.sin_port        = htons(pOptions->port);
    nRet = bind(pRecvInfo->sd[0], (LPSOCKADDR)&addr_srv, sizeof(addr_srv));
    if (SOCKET_ERROR == nRet)
        {
        printf("bind %d", WSAGetLastError());
        return(FALSE);
        }

    if (IPPROTO_TCP == pOptions->spOptions.iProtocol)
        {
        nRet = listen(pRecvInfo->sd[0], SOMAXCONN);
        if (SOCKET_ERROR == nRet)
            {
            printf("listen %d", WSAGetLastError());
            return(FALSE);
            }
        }

    g_sd = pRecvInfo->sd[0];

    // Set the time-to-live (TTL) option for multicasting if enabled.  This limits how
    // far data sent to a group will go.
    if (pOptions->spOptions.bMulticast)
        {
        int    optval = 64;

        if (setsockopt(g_sd, IPPROTO_IP, IP_MULTICAST_TTL, (char *)&optval, sizeof(int)) == SOCKET_ERROR)
            {
            printf("setsockopt(IP_MULTICAST_TTL) %d\n", WSAGetLastError());
	        return(FALSE);
            }
        }

    if (pOptions->spOptions.bQos)
        {
        if (QOS_IOCTL_SET_BEFORE == pOptions->qosOptions.qosIoctlSet)
            {
            if (SetQos(pRecvInfo->sd[0], &pOptions->qosOptions, TRUE, &qos, &cbQosLen))
                printf("  QOS set before accept\n");
            pQos = NULL;
            }
        else if (QOS_IOCTL_SET_QOS == pOptions->qosOptions.qosIoctlSet)
            {
            // this one is tricky.  The intent is to send the RESV after 
            // receiving FD_QOS notification.  However, unless we somehow 
            // inform the GQOSSP that we are interested in QOS, it will not
            // send us FD_QOS when receiving a RESV.  However, we do not want to
            // send out a RESV until we explicitedly set QOS.  We therefore do not 
            // want QOS signalling (i.e. RSVP messages) to be sent.
            pOptions->qosOptions.bDisableSignalling = TRUE;
            if (SetQos(pRecvInfo->sd[0], &pOptions->qosOptions, TRUE, &qos, &cbQosLen))
                {
                printf("  QOS set qos before accept - will be set again in FD_QOS\n");
                pQos = &qos;
                }
            else
                pQos = NULL;
            }
        }
        else
            pQos = NULL;

    // Join the multicast group if specfied.  We may also specify a given QOS structure
    // at this time.
    if (pOptions->spOptions.bMulticast)
        {
        ZeroMemory((PVOID)&mult_addr, sizeof(mult_addr));
        mult_addr.sin_family      = pProtocolInfo->iAddressFamily;
        mult_addr.sin_addr.s_addr = inet_addr(pOptions->szHostname);
        mult_addr.sin_port        = htons(pOptions->port);
        if (WSAJoinLeaf(pRecvInfo->sd[0], (struct sockaddr *)&mult_addr, sizeof(mult_addr), 
                NULL, NULL, pQos, NULL, JL_BOTH) == INVALID_SOCKET)
            {
            printf("WSAJoinLeaf %d", WSAGetLastError());
            return(FALSE);
            }
        }
    return(TRUE);
    }




// Abstract:
//      FD_ACCEPT handler.  Invoked when a client is trying to connect (TCP-only).
//      WSAAccept is called and an Accept Condition function is supplied to allow  
//      a connection to be rejected if too many client have already connected.
//      
//      The condition function allows the accepted socket to be made into a QOS
//      socket (if -i:d option has been specified).  The socket can also be made 
//      into a QOS socket after WSAAccept returns.
//
//      After a successful WSAAccept, the proper notifications for a TCP socket
//      are set with WSAEventSelect.
//
static BOOL fd_accept_func(
    DWORD               dwEvent, 
    WSANETWORKEVENTS    *NetworkEvents, 
    RECV_INFO           *pRecvInfo, 
    OPTIONS             *pOptions
    )
    {
    BOOL    bProcessEventsDone = FALSE;
    SOCKET  temp_socket;
    int     nRet;
    QOS     qos;
    DWORD   cbQosLen;

    
    printf("  FD_ACCEPT: dwEvent=%d error code =%d\n", 
        dwEvent, NetworkEvents->iErrorCode[FD_ACCEPT_BIT]);
    if (NetworkEvents->iErrorCode[FD_ACCEPT_BIT])
        return(TRUE);

    temp_socket = WSAAccept(pRecvInfo->sd[0], NULL, NULL, AcceptCondFunc, (DWORD)pOptions); 
    if (INVALID_SOCKET == temp_socket)
        {
        printf("    WSAAccept %d\n", WSAGetLastError());
        bProcessEventsDone = FALSE;
        }
    else
        {
        // keep list of connected sockets
        printf("    socket connection accepted\n");
        for (int i=1; i<=MAXCLIENTS; i++)
            {
            if (INVALID_SOCKET == pRecvInfo->sd[i])
                {
                pRecvInfo->sd[i] = temp_socket;
                break;
                }
            }

        if ((pOptions->spOptions.bQos) &&
            (QOS_IOCTL_SET_AFTER == pOptions->qosOptions.qosIoctlSet))
            {
            if (SetQos(pRecvInfo->sd[0], &pOptions->qosOptions, TRUE, &qos, &cbQosLen))
                printf("    QOS set after accept\n");
            }

        nRet = WSAEventSelect(temp_socket, pRecvInfo->hEvents[i], (FD_READ|FD_QOS|FD_CLOSE));
        if (SOCKET_ERROR == nRet)
            {
            printf("    WSAEventSelect: %d\n", WSAGetLastError());
            bProcessEventsDone = TRUE;
            }
        }
    
    return(bProcessEventsDone);
    }
    


// Abstract:
//      FD_READ handler.  Handling FD_READ notifications is pretty straight-forward.  
//      Just read the amount of data you are interested in reading at this moment.  
//      If there is more data left, you will get another FD_READ.
//
//      *** Currently, I am reusing the same buffer for all reads.  Though note that 
//      since I am not really "using" any ofthe data I read it is not a problem.
//
static BOOL fd_read_func(
    DWORD               dwEvent, 
    WSANETWORKEVENTS    *NetworkEvents, 
    RECV_INFO           *pRecvInfo, 
    OPTIONS             *pOptions
    )
    {
    BOOL        bProcessEventsDone = FALSE;
    int         nRecv;
    static CHAR *bufptr = pOptions->buf;
    static int  nBytesToRecv = pOptions->nBufSize;

    printf("  FD_READ: dwEvent=%d error code =%d bytestorecv=%d\n", 
        dwEvent, NetworkEvents->iErrorCode[FD_READ_BIT], nBytesToRecv);
    if (NetworkEvents->iErrorCode[FD_READ_BIT])
        return(TRUE);

    nRecv = recv(pRecvInfo->sd[dwEvent], bufptr, nBytesToRecv, 0);
    if (SOCKET_ERROR == nRecv)
        {
        DWORD dwErr= WSAGetLastError();
        if (WSAEWOULDBLOCK != dwErr)
            {
            printf("    recv: %d\n", WSAGetLastError());
            bProcessEventsDone = TRUE;
            }
        else
            {
            printf("    recv: WSAEWOULDBLOCK, wait for next FD_READ\n");
            bProcessEventsDone = FALSE;
            }
        }
    else
        {
        // we got data
        if (nRecv)
            printf("    recv'ed bytes=%d char=<%c>\n", nRecv, *bufptr);
        else
            printf("    recv'ed 0 bytes\n"); 

        bufptr += nRecv;
        nBytesToRecv -= nRecv;
        if (0 == nBytesToRecv)
            {
            bufptr = pOptions->buf;
            nBytesToRecv = pOptions->nBufSize;
            }
        }
    return(bProcessEventsDone);
    }




// Abstract:
//      FD_QOS handler.  This notification indicates that the GQOSSP wants to 
//      indicate a change in QOS status (i.e. delivery of an RSVP message).  
//      Only a change in QOS will cause FD_QOS notifications to be delivered,
//      periodic RSVP refresh messages will not cause an FD_QOS notification.
//
//      Note that the iErrorCode is not really an errorcode but a status code.  
//
//      Also notice that multiple QOS structures may be queued up per FD_QOS 
//      notification and should be examined.
//
static BOOL fd_qos_func(
    DWORD               dwEvent, 
    WSANETWORKEVENTS    *NetworkEvents, 
    RECV_INFO           *pRecvInfo, 
    OPTIONS             *pOptions
    )
    {
    BOOL    bProcessEventsDone = FALSE;
    int     nRet;
    DWORD   cbQosLen;
    DWORD   dwBytes=0;
    DWORD   dwBytesReturned=0;
    QOS     Qos;


    printf("  FD_QOS: dwEvent=%d error/status code =%d\n", 
        dwEvent, NetworkEvents->iErrorCode[FD_QOS_BIT]);

    while (TRUE) 
        {
        if (!GetQos(pRecvInfo->sd[dwEvent], pOptions->qosOptions.bQueryBufferSize, &Qos))
            break;
        else
            {
            PrintQos(&Qos, "    ");
            if ((pOptions->spOptions.bQos) &&
                (QOS_IOCTL_SET_QOS == pOptions->qosOptions.qosIoctlSet))
                {
                // receiving flowspec already has in it what the sender offers/requests
                Qos.SendingFlowspec   = default_notraffic;
            	if (Qos.ProviderSpecific.buf)
                    {
                    free(Qos.ProviderSpecific.buf);
                    Qos.ProviderSpecific.buf = NULL;
                    Qos.ProviderSpecific.len = 0;
                    }
                cbQosLen = sizeof(QOS) + Qos.ProviderSpecific.len;
                nRet = WSAIoctl(pRecvInfo->sd[dwEvent], SIO_SET_QOS, (LPVOID)&Qos, cbQosLen, 
                    NULL, 0, &dwBytes, NULL, NULL);
                if (SOCKET_ERROR == nRet)
                    printf("WSAIoctl:SIO_SET_QOS %d\n", WSAGetLastError());
                else
                    printf("    QOS set during FD_QOS\n");
                }

            if (Qos.ProviderSpecific.buf)
                free(Qos.ProviderSpecific.buf);         
            }
        }
    return(bProcessEventsDone);
    }




// Abstract:
//      FD_CLOSE handler.  When receiving FD_CLOSE, it is still possible for data 
//      to be queued up and therefore the socket should be drained of that data so
//      as not to lose that data.  Winsock2 events are manual reset events and FD_CLOSE 
//      has no re-enabling function, therefore must reset the event here, otherwise 
//      we would get additional FD_CLOSE notifications.
//
static BOOL fd_close_func(
    DWORD               dwEvent, 
    WSANETWORKEVENTS    *NetworkEvents, 
    RECV_INFO           *pRecvInfo, 
    OPTIONS             *pOptions
    )
    {
    BOOL bProcessEventsDone;


    printf("  FD_CLOSE: dwEvent=%d error code =%d\n", 
        dwEvent, NetworkEvents->iErrorCode[FD_CLOSE_BIT]);
    if (NetworkEvents->iErrorCode[FD_CLOSE_BIT])
        bProcessEventsDone = TRUE;
    else
        {
        // drain the socket
        int     nRecv;
        CHAR    *bufptr = pOptions->buf;
        int     nBytesToRecv = pOptions->nBufSize;

        printf("    Draining socket\n");
        while (TRUE)
            {
            nRecv = recv(pRecvInfo->sd[dwEvent], bufptr, nBytesToRecv, 0);
            printf("     recv: nRecv=%d", nRecv);
            if (nRecv > 0) 
                printf(" char=<%c>\n", *bufptr); 
            else
                break;
            }
        printf("\n");

        WSAResetEvent(pRecvInfo->hEvents[dwEvent]);
        closesocket(pRecvInfo->sd[dwEvent]);
        pRecvInfo->sd[dwEvent] = INVALID_SOCKET;
        pOptions->dwTotalClients--;
        bProcessEventsDone = FALSE;
        }
    return(bProcessEventsDone);
    }




// Abstract:
//      Function called by Winsock to accept or reject a connection.  If more 
//      than MAXCLIENTS attempt to connect the connection will be rejected 
//      (CF_REJECT) and the connection reset.  Please note that a connect can 
//      be successful even if the receiving application does not call WSAAccept 
//      (or accept).
//
//      If lpSQos is supplied, then QOS must be set to valid values for Win98 -
//      NT5 does not have this restriction.  On NT5 the contents of lpSQos
//      will be QOS_NOT_SPECIFIED (even if PATH data has arrived prior to 
//      executing this function).  This behaviour is by design for the 
//      Microsoft GQOSSP.
//
int CALLBACK AcceptCondFunc(
    LPWSABUF    lpCallerId,
    LPWSABUF    lpCallerData,
    LPQOS       lpSQos,
    LPQOS       lpGQos,
    LPWSABUF    lpCalleeId,
    LPWSABUF    lpCalleeData,
    GROUP FAR   *Group,
    DWORD       CallbackData
    )
    {
    int     nRet        = CF_ACCEPT;
    OPTIONS *pOptions   = (OPTIONS *)CallbackData;


    printf("    AcceptCondFunc\n");
    if (lpCallerId)
        {
        // NOTE:  This program only supports TCP or UDP
        SOCKADDR_IN *sockaddr = (SOCKADDR_IN *)lpCallerId->buf;
        printf("      lpCallerId address family=%d, port=%d\n", 
            sockaddr->sin_family, ntohs(sockaddr->sin_port));
        }

    if (lpSQos)
        {
        printf("    AcceptCondFunc: lpSQos valid\n");
        if (pOptions->spOptions.bQos)
            {
            // bMustSetQosInAccept will not be set for NT5 but only for Win98.
            // Win98 currently requires valid QOS parameters to be set here iff
            // lpSQos is a valid pointer.
            if ((QOS_IOCTL_SET_DURING == pOptions->qosOptions.qosIoctlSet) ||
                pOptions->qosOptions.bMustSetQosInAccept)
                {
                DWORD cbQosLen;
                printf("    AcceptCondFunc: filling in Qos\n");
                SetQos(g_sd, &pOptions->qosOptions,  FALSE, lpSQos, &cbQosLen);
                PrintQos(lpSQos, "    ");
                }
            }
        }

    if (lpCalleeId)
        {
        // NOTE:  This program only supports TCP or UDP
        SOCKADDR_IN *sockaddr = (SOCKADDR_IN *)lpCalleeId->buf;
        printf("      lpCalleeId: address family=%d, port=%d\n", 
            sockaddr->sin_family, ntohs(sockaddr->sin_port));
        }

    if (MAXCLIENTS == pOptions->dwTotalClients)
        {
        printf("      Rejecting connection, limit of %d clients reached\n",
            pOptions->dwTotalClients);
        nRet = CF_REJECT;
        // At this point a connection will already have been established,
        // this will just cause the connection to be reset.
        }
    else
        pOptions->dwTotalClients++;

    return(nRet);
    } 

